@@ -12,6 +12,8 @@ from typing import (
     get_type_hints,
     Callable,
     Generic,
+    Hashable,
+    ClassVar,
 )
 import re
 from datetime import datetime
@@ -35,6 +37,8 @@ JSON_ENCODABLE_TYPES = {
 JsonEncodable = Union[int, float, str, bool, None]
 JsonDict = Dict[str, Any]
 
+OPTIONAL_TYPES = ["Union", "Optional"]
+
 
 class ValidationError(jsonschema.ValidationError):
     pass
@@ -83,7 +87,9 @@ def issubclass_safe(klass: Any, base: Type) -> bool:
 
 
 def is_optional(field: Any) -> bool:
-    if str(field).startswith("typing.Union"):
+    if str(field).startswith("typing.Union") or str(field).startswith(
+        "typing.Optional"
+    ):
         for arg in field.__args__:
             if isinstance(arg, type) and issubclass(arg, type(None)):
                 return True
@@ -179,7 +185,8 @@ class FieldMeta:
 
 
 @functools.lru_cache()
-def _validate_schema(schema_cls: Type[T]) -> JsonDict:
+def _validate_schema(h_schema_cls: Hashable) -> JsonDict:
+    schema_cls = cast(Type[JsonSchemaMixin], h_schema_cls)  # making mypy happy
     schema = schema_cls.json_schema()
     jsonschema.Draft7Validator.check_schema(schema)
     return schema
@@ -276,20 +283,22 @@ class JsonSchemaMixin:
     convert to and from JSON encodable dicts with validation against the schema
     """
 
-    _field_encoders: Dict[Type, FieldEncoder] = {
+    _field_encoders: ClassVar[Dict[Type, FieldEncoder]] = {
         datetime: DateTimeFieldEncoder(),
         UUID: UuidField(),
     }
 
     # Cache of the generated schema
-    _schema: Optional[Dict[str, CompleteSchema]] = None
+    _schema: ClassVar[Optional[Dict[str, CompleteSchema]]] = None
 
     # Cache of field encode / decode functions
-    _encode_cache: Optional[Dict[Any, _ValueEncoder]] = None
-    _decode_cache: Optional[Dict[Any, _ValueDecoder]] = None
-    _mapped_fields: Optional[List[Tuple[Field, str]]] = None
+    _encode_cache: ClassVar[Optional[Dict[Any, _ValueEncoder]]] = None
+    _decode_cache: ClassVar[Optional[Dict[Any, _ValueDecoder]]] = None
+    _mapped_fields: ClassVar[
+        Optional[Dict[Any, List[Tuple[Field, str]]]]
+    ] = None
 
-    ADDITIONAL_PROPERTIES = False
+    ADDITIONAL_PROPERTIES: ClassVar[bool] = False
 
     @classmethod
     def field_mapping(cls) -> Dict[str, str]:
@@ -310,6 +319,9 @@ class JsonSchemaMixin:
         else:
             cls._field_encoders.update(field_encoders)
 
+    def _local_to_dict(self, **kwargs):
+        return self.to_dict(**kwargs)
+
     @classmethod
     def _encode_field(
         cls, field_type: Any, value: Any, omit_none: bool
@@ -333,7 +345,7 @@ class JsonSchemaMixin:
                 def encoder(_, v, __):
                     return v.value
 
-            elif field_type_name == "Union":
+            elif field_type_name in OPTIONAL_TYPES:
                 # Attempt to encode the field with each union variant.
                 # TODO: Find a more reliable method than this since in the case 'Union[List[str], Dict[str, int]]' this
                 # will just output the dict keys as a list
@@ -410,7 +422,9 @@ class JsonSchemaMixin:
             elif cls._is_json_schema_subclass(field_type):
                 # Only need to validate at the top level
                 def encoder(_, v, o):
-                    return v.to_dict(omit_none=o, validate=False)
+                    # this calls _local_to_dict in order to support
+                    # combining this code with mashumaro
+                    return v._local_to_dict(omit_none=o)
 
             elif hasattr(field_type, "__supertype__"):  # NewType field
 
@@ -427,20 +441,33 @@ class JsonSchemaMixin:
 
     @classmethod
     def _get_fields(cls) -> List[Tuple[Field, str]]:
-        mapped_fields = []
-        type_hints = get_type_hints(cls)
+        if cls._mapped_fields is None:
+            cls._mapped_fields = {}
+        if cls.__name__ not in cls._mapped_fields:
+            mapped_fields = []
+            type_hints = get_type_hints(cls)
 
-        for f in fields(cls):
-            # Skip internal fields
-            if f.name.startswith("_"):
-                continue
+            for f in fields(cls):
+                # Skip internal fields
+                if f.name.startswith("_"):
+                    continue
 
-            # Note fields() doesn't resolve forward refs
-            f.type = type_hints[f.name]
+                # Note fields() doesn't resolve forward refs
+                f.type = type_hints[f.name]
 
-            mapped_fields.append((f, cls.field_mapping().get(f.name, f.name)))
+                mapped_fields.append(
+                    (f, cls.field_mapping().get(f.name, f.name))
+                )
+            cls._mapped_fields[cls.__name__] = mapped_fields
+        return cls._mapped_fields[cls.__name__]
 
-        return mapped_fields  # type: ignore
+    @classmethod
+    def _get_field_names(cls):
+        fields = cls._get_fields()
+        field_names = []
+        for element in fields:
+            field_names.append(element[1])
+        return field_names
 
     def to_dict(
         self, omit_none: bool = True, validate: bool = False
@@ -491,7 +518,7 @@ class JsonSchemaMixin:
                 def decoder(_, ft, val):
                     return ft.from_dict(val, validate=validate)
 
-            elif field_type_name == "Union":
+            elif field_type_name in OPTIONAL_TYPES:
                 # Attempt to decode the value using each decoder in turn
                 union_excs = (
                     AttributeError,
@@ -725,7 +752,7 @@ class JsonSchemaMixin:
             field_schema.update(cls._encode_restrictions(restrictions))
 
         # if Union[..., None] or Optional[...]
-        elif type_name == "Union":
+        elif type_name in OPTIONAL_TYPES:
             field_schema = {
                 "oneOf": [
                     cls._get_field_schema(variant)[0]
@@ -828,7 +855,7 @@ class JsonSchemaMixin:
             cls._get_field_definitions(field_type.__args__[1], definitions)
         elif field_type_name == "PatternProperty":
             cls._get_field_definitions(field_type.TARGET_TYPE, definitions)
-        elif field_type_name == "Union":
+        elif field_type_name in OPTIONAL_TYPES:
             for variant in field_type.__args__:
                 cls._get_field_definitions(variant, definitions)
         elif cls._is_json_schema_subclass(field_type):
@@ -954,7 +981,8 @@ class JsonSchemaMixin:
 
     @classmethod
     def validate(cls, data: Any):
-        schema = _validate_schema(cls)
+        h_cls = cast(Hashable, cls)
+        schema = _validate_schema(h_cls)
         validator = jsonschema.Draft7Validator(schema)
         error = next(iter(validator.iter_errors(data)), None)
         if error is not None:
